/*
 * PIC32 peripheral configuration and initialisation.
 *
 * Copyright (C) 2017, 2018 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "cpu.h"
#include "pic32_c.h"
#include "init.h"



/* Basic memory and pin initialisation. */

void init_memory(void)
{
    /*
    Configure RAM.
    See: http://microchipdeveloper.com/32bit:mx-arch-exceptions-processor-initialization
    */

    uint32_t config = REG(BMXCON);

    /* Set zero wait states for address setup. */

    config &= ~(1 << 6);        /* BMXCON<6> = BMXWSDRM = 0 */

    /* Set bus arbitration mode. */

    config &= ~0b111;
    config |= 0b010;            /* BMXCON<2:0> = BMXARB<2:0> = 2 */

    REG(BMXCON) = config;
}

void init_pins(void)
{
    /* DEVCFG0<2> also needs setting to 0 before the program is run. */

    CLR_REG(CFGCON, 1 << 3);    /* CFGCON<3> = JTAGEN = 0 */
}

void init_outputs(void)
{
    /* Remove analogue features from pins. */

    REG(ANSELA) = 0;
    REG(ANSELB) = 0;

    /* Set pins as outputs. */

    REG(TRISA) = 0;
    REG(TRISB) = 0;

    /* Clear outputs. */

    REG(PORTA) = 0;
    REG(PORTB) = 0;
}



/* Peripheral pin configuration. */

void lock_config(void)
{
    SET_REG(CFGCON, 1 << 13);   /* IOLOCK = 1 */

    /* Lock the configuration again. */

    REG(SYSKEY) = 0x33333333;
}

void unlock_config(void)
{
    /* Unlock the configuration register bits. */

    REG(SYSKEY) = 0;
    REG(SYSKEY) = 0xAA996655;
    REG(SYSKEY) = 0x556699AA;

    CLR_REG(CFGCON, 1 << 13);   /* IOLOCK = 0 */
}



/* Convenience operations. */

void interrupts_on(void)
{
    init_interrupts();
    enable_interrupts();
    handle_error_level();
}



/* DMA configuration. */

void init_dma(void)
{
    /* Disable DMA interrupts (DMAxIE). */

    CLR_REG(DMAIEC, 0b1111 << DMAINTBASE);

    /* Clear DMA interrupt flags (DMAxIF). */

    CLR_REG(DMAIFS, 0b1111 << DMAINTBASE);

    /* Enable DMA. */

    SET_REG(DMACON, 1 << 15);
}

/* Initialise the given channel. */

void dma_init(int channel, uint8_t pri)
{
    if ((channel < DCHMIN) || (channel > DCHMAX))
        return;

    /* Initialise a channel. */

    REG(DMA_REG(channel, DCHxCON)) = pri & 0b11;
    REG(DMA_REG(channel, DCHxECON)) = 0;
    REG(DMA_REG(channel, DCHxINT)) = 0;
}

/* Set the channel repeated enable mode, enabling it again when a block transfer
   completes. The documentation describes this as auto-enable. */

void dma_set_auto_enable(int channel, int enable)
{
    (enable ? SET_REG : CLR_REG)(DMA_REG(channel, DCHxCON), 1 << 4);
}

/* Set the channel chaining mode. */

void dma_set_chaining(int channel, enum dma_chain chain)
{
    (chain != dma_chain_none ?
     SET_REG : CLR_REG)(DMA_REG(channel, DCHxCON), 1 << 5);

    (chain == dma_chain_next ?
     SET_REG : CLR_REG)(DMA_REG(channel, DCHxCON), 1 << 8);
}

/* Configure a channel's initiation interrupt. */

void dma_set_interrupt(int channel, uint8_t int_num, int enable)
{
    if ((channel < DCHMIN) || (channel > DCHMAX))
        return;

    /* Allow an interrupt to trigger the transfer. */

    REG(DMA_REG(channel, DCHxECON)) = (int_num << 8) |
                                      ((enable ? 1 : 0) << 4);
}

/* Configure only the channel's initiation interrupt status. */

void dma_set_interrupt_enable(int channel, int enable)
{
    if ((channel < DCHMIN) || (channel > DCHMAX))
        return;

    (enable ? SET_REG : CLR_REG)(DMA_REG(channel, DCHxECON), 1 << 4);
}

/* Permit the channel to register events while disabled or suspended. A
   suspended channel is one that is enabled but where the DMA peripheral
   has been suspended. */

void dma_set_receive_events(int channel, int enable)
{
    (enable ? SET_REG : CLR_REG)(DMA_REG(channel, DCHxCON), 1 << 6);
}

/* Set a channel's source. */

void dma_set_source(int channel,
                    uint32_t source_start_address, uint16_t source_size)
{
    if ((channel < DCHMIN) || (channel > DCHMAX))
        return;

    REG(DMA_REG(channel, DCHxSSIZ)) = source_size;
    REG(DMA_REG(channel, DCHxSSA)) = source_start_address;
}

/* Set a channel's transfer parameters. */

void dma_set_transfer(int channel,
                  uint32_t source_start_address, uint16_t source_size,
                  uint32_t destination_start_address, uint16_t destination_size,
                  uint16_t cell_size)
{
    if ((channel < DCHMIN) || (channel > DCHMAX))
        return;

    dma_set_source(channel, source_start_address, source_size);

    REG(DMA_REG(channel, DCHxDSIZ)) = destination_size;
    REG(DMA_REG(channel, DCHxDSA)) = destination_start_address;
    REG(DMA_REG(channel, DCHxCSIZ)) = cell_size;
}

/* Configure interrupts caused by the channel. */

void dma_init_interrupt(int channel, uint8_t conditions,
                        uint8_t pri, uint8_t sub)
{
    if ((channel < DCHMIN) || (channel > DCHMAX))
        return;

    /* Disable channel interrupt and clear interrupt flag. */

    CLR_REG(DMAIEC, DMA_INT_FLAGS(channel, 1));
    CLR_REG(DMAIFS, DMA_INT_FLAGS(channel, 1));

    /* Produce an interrupt for the provided conditions. */

    REG(DMA_REG(channel, DCHxINT)) = conditions << 16;

    /* Set interrupt priorities. */

    REG(DMAIPC) = (REG(DMAIPC) &
                   ~(DMA_IPC_PRI(channel, 7, 3))) |
                  DMA_IPC_PRI(channel, pri, sub);

    /* Enable interrupt. */

    SET_REG(DMAIEC, DMA_INT_FLAGS(channel, 1));
}

/* Enable or disable the channel. */

void dma_set_enable(int channel, int enable)
{
    if ((channel < DCHMIN) || (channel > DCHMAX))
        return;

    (enable ? SET_REG : CLR_REG)(DMA_REG(channel, DCHxCON), 1 << 7);
}

/* Disable a DMA channel. */

void dma_off(int channel)
{
    dma_set_enable(channel, 0);
}

/* Enable a DMA channel. */

void dma_on(int channel)
{
    dma_set_enable(channel, 1);
}



/* External interrupt initialisation. */

void int_init_interrupt(int int_num, uint8_t pri, uint8_t sub)
{
    if ((int_num < INTMIN) || (int_num > INTMAX))
        return;

    /* Disable interrupt and clear interrupt flag. */

    CLR_REG(INTIEC, INT_INT_FLAGS(int_num, INTxIE));
    CLR_REG(INTIFS, INT_INT_FLAGS(int_num, INTxIF));

    /* Set interrupt priorities. */

    REG(INT_IPC_REG(int_num)) = (REG(INT_IPC_REG(int_num)) &
                                 ~(INT_IPC_PRI(int_num, 7, 3))) |
                                INT_IPC_PRI(int_num, pri, sub);

    /* Enable interrupt. */

    SET_REG(INTIEC, INT_INT_FLAGS(int_num, INTxIE));
}



/* Output compare configuration. */

void oc_init(int unit, uint8_t mode, int timer)
{
    if ((unit < OCMIN) || (unit > OCMAX))
        return;

    REG(OC_REG(unit, OCxCON)) = (timer == 3 ? (1 << 3) : 0) | (mode & 0b111);
}

/* Set the start value for the pulse. */

void oc_set_pulse(int unit, uint32_t start)
{
    if ((unit < OCMIN) || (unit > OCMAX))
        return;

    REG(OC_REG(unit, OCxR)) = start;
}

/* Set the end value for the pulse. */

void oc_set_pulse_end(int unit, uint32_t end)
{
    if ((unit < OCMIN) || (unit > OCMAX))
        return;

    REG(OC_REG(unit, OCxRS)) = end;
}

/* Configure interrupts caused by the unit. */

void oc_init_interrupt(int unit, uint8_t pri, uint8_t sub)
{
    if ((unit < OCMIN) || (unit > OCMAX))
        return;

    /* Disable interrupt and clear interrupt flag. */

    CLR_REG(OCIEC, OC_INT_FLAGS(unit, OCxIE));
    CLR_REG(OCIFS, OC_INT_FLAGS(unit, OCxIF));

    /* Set interrupt priorities. */

    REG(OC_IPC_REG(unit)) = (REG(OC_IPC_REG(unit)) &
                             ~(OC_IPC_PRI(unit, 7, 3))) |
                            OC_IPC_PRI(unit, pri, sub);

    /* Enable interrupt. */

    SET_REG(OCIEC, OC_INT_FLAGS(unit, OCxIE));
}

/* Enable a unit. */

void oc_on(int unit)
{
    if ((unit < OCMIN) || (unit > OCMAX))
        return;

    SET_REG(OC_REG(unit, OCxCON), 1 << 15);
}



/* Parallel mode configuration. */

void init_pm(void)
{
    int i;

    /* Disable PM interrupts (PMxIE). */

    CLR_REG(PMIEC, 0b11 << PMINTBASE);

    /* Clear PM interrupt flags (PMxIF). */

    CLR_REG(PMIFS, 0b11 << PMINTBASE);

    /* Disable PM for configuration. */

    for (i = PMMIN; i <= PMMAX; i++)
        REG(PM_REG(i, PMxCON)) = 0;
}

/* Configure the parallel mode. */

void pm_init(int port, uint8_t mode)
{
    if ((port < PMMIN) || (port > PMMAX))
        return;

    REG(PM_REG(port, PMxMODE)) = (mode & 0b11) << 8;
    REG(PM_REG(port, PMxAEN)) = 0;
    REG(PM_REG(port, PMxADDR)) = 0;
}

/* Configure output signals. */

void pm_set_output(int port, int write_enable, int read_enable)
{
    if ((port < PMMIN) || (port > PMMAX))
        return;

    REG(PM_REG(port, PMxCON)) = (write_enable ? (1 << 9) : 0) |
                                (read_enable ? (1 << 8) : 0) |
                                (1 << 1); /* WRSP: PMENB active high */
}

/* Configure interrupts caused by parallel mode. */

void pm_init_interrupt(int port, uint8_t pri, uint8_t sub)
{
    if ((port < PMMIN) || (port > PMMAX))
        return;

    /* Disable interrupt and clear interrupt flag. */

    CLR_REG(PMIEC, PM_INT_FLAGS(port, PMxIE));
    CLR_REG(PMIFS, PM_INT_FLAGS(port, PMxIF));

    /* Set interrupt priorities. */

    REG(PM_IPC_REG(port)) = (REG(PM_IPC_REG(port)) &
                             ~(PM_IPC_PRI(port, 7, 3))) |
                            PM_IPC_PRI(port, pri, sub);

    /* Enable interrupt. */

    SET_REG(PMIEC, PM_INT_FLAGS(port, PMxIE));
}

/* Enable parallel mode. */

void pm_on(int port)
{
    if ((port < PMMIN) || (port > PMMAX))
        return;

    SET_REG(PM_REG(port, PMxCON), 1 << 15);
}

/* Disable parallel mode. */

void pm_off(int port)
{
    if ((port < PMMIN) || (port > PMMAX))
        return;

    CLR_REG(PM_REG(port, PMxCON), 1 << 15);
}




/* Timer configuration. */

void timer_init(int timer, uint8_t prescale, uint16_t limit)
{
    /* NOTE: Should convert from the real prescale value. */

    REG(TIMER_REG(timer, TxCON)) = (prescale & 0b111) << 4;
    REG(TIMER_REG(timer, TMRx)) = 0;
    REG(TIMER_REG(timer, PRx)) = limit;
}

/* Configure interrupts caused by the timer. */

void timer_init_interrupt(int timer, uint8_t pri, uint8_t sub)
{
    if ((timer < TIMERMIN) || (timer > TIMERMAX))
        return;

    /* Disable interrupt and clear interrupt flag. */

    CLR_REG(TIMERIEC, TIMER_INT_FLAGS(timer, TxIE));
    CLR_REG(TIMERIFS, TIMER_INT_FLAGS(timer, TxIF));

    /* Set interrupt priorities. */

    REG(TIMER_IPC_REG(timer)) = (REG(TIMER_IPC_REG(timer)) &
                                 ~(TIMER_IPC_PRI(timer, 7, 3))) |
                                TIMER_IPC_PRI(timer, pri, sub);

    /* Enable interrupt. */

    SET_REG(TIMERIEC, TIMER_INT_FLAGS(timer, TxIE));
}

/* Enable a timer. */

void timer_on(int timer)
{
    if ((timer < TIMERMIN) || (timer > TIMERMAX))
        return;

    SET_REG(TIMER_REG(timer, TxCON), 1 << 15);
}



/* UART configuration. */

void uart_init(int uart, int fpb, uint32_t baudrate)
{
    /* FPB is configured in the devconfig.h file and set in the start.S file. */

    if ((uart < UARTMIN) || (uart > UARTMAX))
        return;

    /* Disable the UART (ON). */

    CLR_REG(UART_REG(uart, UxMODE), 1 << 15);

    /* Set the baud rate. For example:

       UxBRG<15:0> = BRG
                   = (FPB / (16 * baudrate)) - 1
                   = (24000000 / (16 * 115200)) - 1
                   = 12
    */

    REG(UART_REG(uart, UxBRG)) = (fpb / (16 * baudrate)) - 1;
}

/* Configure interrupts caused by the UART. */

void uart_init_interrupt(int uart, uint8_t conditions,
                         uint8_t pri, uint8_t sub)
{
    if ((uart < UARTMIN) || (uart > UARTMAX))
        return;

    /* Disable interrupts and clear interrupt flags. */

    CLR_REG(UARTIEC, UART_INT_FLAGS(uart, UxTIE | UxRIE | UxEIE));
    CLR_REG(UARTIFS, UART_INT_FLAGS(uart, UxTIF | UxRIF | UxEIF));

    /* Set priorities: UxIP = pri; UxIS = sub */

    REG(UART_IPC_REG(uart)) = (REG(UART_IPC_REG(uart)) &
                               ~UART_IPC_PRI(uart, 7, 3)) |
                              UART_IPC_PRI(uart, pri, sub);

    /* Enable interrupts. */

    SET_REG(UARTIEC, UART_INT_FLAGS(uart, conditions));
}

/* Enable a UART. */

void uart_on(int uart)
{
    if ((uart < UARTMIN) || (uart > UARTMAX))
        return;

    /* Enable receive (URXEN) and transmit (UTXEN). */

    SET_REG(UART_REG(uart, UxSTA), (1 << 12) | (1 << 10));

    /* Start UART. */

    SET_REG(UART_REG(uart, UxMODE), 1 << 15);
}



/* Utility functions. */

/* Return encoded interrupt priorities. */

static uint8_t PRI(uint8_t pri, uint8_t sub)
{
    return ((pri & 0b111) << 2) | (sub & 0b11);
}

/* Return the DMA interrupt flags for combining with a register. */

int DMA_INT_FLAGS(int channel, uint8_t flags)
{
    return (flags & 0b1) << (DMAINTBASE + (channel - DCHMIN));
}

/* Return encoded DMA interrupt priorities for combining with a register. */

uint32_t DMA_IPC_PRI(int channel, uint8_t pri, uint8_t sub)
{
    return PRI(pri, sub) << (DCHIPCBASE + (channel - DCHMIN) * DCHIPCSTEP);
}

/* Return encoded external interrupt priorities for combining with a register. */

uint32_t INT_IPC_PRI(int int_num, uint8_t pri, uint8_t sub)
{
    (void) int_num;
    return PRI(pri, sub) << INTIPCBASE;
}

/* Return the external interrupt priorities register. */

uint32_t INT_IPC_REG(int int_num)
{
    switch (int_num)
    {
        case 0: return INT0IPC;
        case 1: return INT1IPC;
        case 2: return INT2IPC;
        case 3: return INT3IPC;
        case 4: return INT4IPC;
        default: return 0; /* should not occur */
    }
}

/* Return the external interrupt flags for combining with a register. */

int INT_INT_FLAGS(int int_num, uint8_t flags)
{
    return (flags & 0b1) << (INTINTBASE + (int_num - INTMIN) * INTINTSTEP);
}

/* Return encoded output compare interrupt priorities for combining with a register. */

uint32_t OC_IPC_PRI(int unit, uint8_t pri, uint8_t sub)
{
    (void) unit;
    return PRI(pri, sub) << OCIPCBASE;
}

/* Return the output compare interrupt priorities register. */

uint32_t OC_IPC_REG(int unit)
{
    switch (unit)
    {
        case 1: return OC1IPC;
        case 2: return OC2IPC;
        case 3: return OC3IPC;
        case 4: return OC4IPC;
        case 5: return OC5IPC;
        default: return 0; /* should not occur */
    }
}

/* Return the output compare interrupt flags for combining with a register. */

int OC_INT_FLAGS(int unit, uint8_t flags)
{
    return (flags & 0b1) << (OCINTBASE + (unit - OCMIN) * OCINTSTEP);
}

/* Return encoded parallel mode interrupt priorities for combining with a register. */

uint32_t PM_IPC_PRI(int port, uint8_t pri, uint8_t sub)
{
    (void) port;
    return PRI(pri, sub) << PMIPCBASE;
}

/* Return the parallel mode interrupt priorities register. */

uint32_t PM_IPC_REG(int port)
{
    (void) port;
    return PMIPC;
}

/* Return the parallel mode interrupt flags for combining with a register. */

int PM_INT_FLAGS(int port, uint8_t flags)
{
    return (flags & 0b11) << (PMINTBASE + (port - PMMIN) * PMINTSTEP);
}

/* Return encoded timer interrupt priorities for combining with a register. */

uint32_t TIMER_IPC_PRI(int timer, uint8_t pri, uint8_t sub)
{
    (void) timer;
    return PRI(pri, sub) << TIMERIPCBASE;
}

/* Return the timer interrupt priorities register. */

uint32_t TIMER_IPC_REG(int timer)
{
    switch (timer)
    {
        case 1: return TIMER1IPC;
        case 2: return TIMER2IPC;
        case 3: return TIMER3IPC;
        case 4: return TIMER4IPC;
        case 5: return TIMER5IPC;
        default: return 0; /* should not occur */
    }
}

/* Return the timer interrupt flags for combining with a register. */

int TIMER_INT_FLAGS(int timer, uint8_t flags)
{
    return (flags & 0b1) << (TIMERINTBASE + (timer - TIMERMIN) * TIMERINTSTEP);
}

/* Return encoded UART interrupt priorities for combining with a register. */

uint32_t UART_IPC_PRI(int uart, uint8_t pri, uint8_t sub)
{
    return PRI(pri, sub) << (uart == 1 ? UART1IPCBASE : UART2IPCBASE);
}

/* Return the UART interrupt priorities register. */

uint32_t UART_IPC_REG(int uart)
{
    return uart == 1 ? UART1IPC : UART2IPC;
}

/* Return the UART interrupt flags for combining with a register. */

int UART_INT_FLAGS(int uart, uint8_t flags)
{
    return (flags & 0b111) << (UARTINTBASE + (uart - UARTMIN) * UARTINTSTEP);
}
